from __future__ import annotations

import numpy as np

from rq_interfaces.msg import ProjectionGeometry, Projection
from scipy.spatial.transform import Rotation

class PyProjectionGeometry():
    """
    Represents the geometry of a projection including the focal spot and detector position and orientation.

    Attributes:
        focal_spot_mm (np.ndarray): Position of the focal spot in millimeters.
        detector_postion_mm (np.ndarray): Position of the detector in millimeters.
        detector_orientation_quad (np.ndarray): Orientation of the detector as a quaternion.
        focal_spot_orientation_quad (np.ndarray): Orientation of the focal spot as a quaternion.
        frame_id (str): Frame ID for the projection geometry.
    """

    def __init__(self, focal_spot_mm: np.ndarray, detector_postion_mm: np.ndarray, 
                 detector_orientation_quad: np.ndarray,
                 frame_id: str = 'object', focal_spot_orientation_quad: np.ndarray = np.array([0., 0., 0, 1.])
                 ) -> None:
        """
        Initializes a PyProjectionGeometry instance.

        Args:
            focal_spot_mm (np.ndarray): Position of the focal spot in millimeters.
            detector_postion_mm (np.ndarray): Position of the detector in millimeters.
            detector_orientation_quad (np.ndarray): Orientation of the detector as a quaternion.
            frame_id (str): Frame ID for the projection geometry. Default is 'object'.
            focal_spot_orientation_quad (np.ndarray): Orientation of the focal spot as a quaternion.
        """

        self.focal_spot_mm = focal_spot_mm
        self.detector_postion_mm = detector_postion_mm
        self.focal_spot_orientation_quad = focal_spot_orientation_quad
        self.detector_orientation_quad = detector_orientation_quad
        self.frame_id = frame_id

    @classmethod
    def dummy(cls):
        """
        Creates a dummy instance of PyProjectionGeometry for testing.

        Returns:
            PyProjectionGeometry: A dummy instance with default values.
        """
        return cls(np.array([1., 0., 0]), 
                   np.array([-1., 0., 0]),
                   np.array([0., 0., 0, 1.]))

    @classmethod
    def from_message(cls, msg: ProjectionGeometry):
        """
        Creates an instance of PyProjectionGeometry from a ROS message.

        Args:
            msg (ProjectionGeometry): The ROS message containing projection geometry data.

        Returns:
            PyProjectionGeometry: An instance initialized from the ROS message.
        """
        focal_spot_mm = np.array([msg.focal_spot_postion_mm.x,
                                  msg.focal_spot_postion_mm.y,
                                  msg.focal_spot_postion_mm.z,])

        detector_center_mm = np.array([msg.detector_postion_mm.x,
                                       msg.detector_postion_mm.y,
                                       msg.detector_postion_mm.z,])
        
        detector_orientation_quad = np.array([msg.detector_orientation_quad.x,
                                              msg.detector_orientation_quad.y,
                                              msg.detector_orientation_quad.z,
                                              msg.detector_orientation_quad.w])
        
        focal_spot_orientation = np.array([msg.focal_spot_orientation_quad.x,
                                           msg.focal_spot_orientation_quad.y,
                                           msg.focal_spot_orientation_quad.z,
                                           msg.focal_spot_orientation_quad.w])
        
        frame_id = msg.header.frame_id

        return cls(focal_spot_mm, detector_center_mm, detector_orientation_quad, frame_id, focal_spot_orientation)
    
    def as_message(self) -> ProjectionGeometry:
        """
        Converts the PyProjectionGeometry instance to a ROS message.

        Returns:
            ProjectionGeometry: The ROS message representing the projection geometry.
        """
        message = ProjectionGeometry()

        message.focal_spot_postion_mm.x = self.focal_spot_mm[0]
        message.focal_spot_postion_mm.y = self.focal_spot_mm[1]
        message.focal_spot_postion_mm.z = self.focal_spot_mm[2]

        message.detector_postion_mm.x = self.detector_postion_mm[0]
        message.detector_postion_mm.y = self.detector_postion_mm[1]
        message.detector_postion_mm.z = self.detector_postion_mm[2]

        message.focal_spot_orientation_quad.x = self.focal_spot_orientation_quad[0]
        message.focal_spot_orientation_quad.y = self.focal_spot_orientation_quad[1]
        message.focal_spot_orientation_quad.z = self.focal_spot_orientation_quad[2]
        message.focal_spot_orientation_quad.w = self.focal_spot_orientation_quad[3]

        message.detector_orientation_quad.x = self.detector_orientation_quad[0]
        message.detector_orientation_quad.y = self.detector_orientation_quad[1]
        message.detector_orientation_quad.z = self.detector_orientation_quad[2]
        message.detector_orientation_quad.w = self.detector_orientation_quad[3]

        message.header.frame_id = self.frame_id

        return message
    
    @property
    def detector_rotation_matrix(self) -> np.ndarray:
        return Rotation.from_quat(self.detector_orientation_quad).as_matrix()

    
    @property 
    def detector_horizontal_vector(self) -> np.ndarray:
        return self.detector_rotation_matrix[:, 0]
    
    @property 
    def detector_vertical_vector(self) -> np.ndarray:
        return self.detector_rotation_matrix[:, 1]

    
    @property
    def projection_matrix(self) -> np.ndarray:
        p3x3 = np.vstack([self.detector_horizontal_vector,
                          self.detector_vertical_vector, 
                          self.detector_postion_mm - self.focal_spot_mm]).T
        p3x3_inv = np.linalg.inv(p3x3)
        p4 = (p3x3_inv @ (-self.focal_spot_mm)).reshape((3, 1))
        matrix = np.concatenate([p3x3_inv, p4], 1).reshape((3, 4))
        return matrix
